/* 
 * File:   lab1.c
 * Author: avsej
 *
 * Created on June 5, 2009, 8:59 AM
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include "pci_codes.h"

#define	CONFIG_ADDRESS	0x0CF8
#define	CONFIG_DATA	0x0CFC

int main(int argc, char **argv) {
    if (iopl(3)) {
        perror("iopl");
        return 1;
    }

    int i;
    int bus;
    for (bus = 0; bus < (1 << 8); bus++) {
        int dev;
        for (dev = 0; dev < (1 << 5); dev++) {
            int busdev = (bus << 16) | (dev << 11) | (1 << 31);
            int pcidat, vendor_id, device_id, class_id;

            outl(busdev, CONFIG_ADDRESS);
            pcidat = inl(CONFIG_DATA);
            if (pcidat == ~0) continue;

            vendor_id = (pcidat >> 0)&0xFFFF;
            device_id = (pcidat >> 16)&0xFFFF;

            outl(busdev | (1 << 3), CONFIG_ADDRESS);
            class_id = inl(CONFIG_DATA) >> 8;

            printf("bus=%-3X dev=%-2X", bus, dev);
            const char* vendor_name = NULL;
            for (i = 0; i < PCI_VENTABLE_LEN; i++) {
                if (vendor_id == PciVenTable[i].VenId) {
                    vendor_name = PciVenTable[i].VenFull;
                    break;
                }
            }
            const char* device_name = NULL;
            const char* device_desc = NULL;
            for (i = 0; i < PCI_DEVTABLE_LEN; i++) {
                if (device_id == PciDevTable[i].DevId && vendor_id == PciDevTable[i].VenId) {
                    device_name = PciDevTable[i].Chip;
                    device_desc = PciDevTable[i].ChipDesc;
                    break;
                }
            }
            const char* class_base_desc = NULL;
            const char* class_sub_desc = NULL;
            const char* class_prog_desc = NULL;
            for (i = 0; i < PCI_CLASSCODETABLE_LEN; i++) {
                if (((class_id >> 16) & 0xFF) == PciClassCodeTable[i].BaseClass && 
                    ((class_id >> 8) & 0xFF) == PciClassCodeTable[i].SubClass &&
                    ((class_id) & 0xFF) == PciClassCodeTable[i].ProgIf
                    ) {
                    class_base_desc = PciClassCodeTable[i].BaseDesc;
                    class_sub_desc = PciClassCodeTable[i].SubDesc;
                    class_prog_desc = PciClassCodeTable[i].ProgDesc;
                    break;
                }
            }

            printf(" VENDOR=%04X ", vendor_id);
            if (vendor_name) {
                printf("(%s) ", vendor_name);
            }
            printf(" DEVICE=%04X ", device_id);
            if (device_desc && device_name) {
                printf("(%s %s) ", device_desc, device_name);
            }
            printf(" CLASS=%02X %02X %02X ", ((class_id >> 16) & 0xFF), ((class_id >> 8) & 0xFF), ((class_id) & 0xFF));
            if (class_base_desc && class_sub_desc && class_prog_desc) {
                printf("(%s; %s; %s) ", class_base_desc, class_sub_desc, class_prog_desc);
            }
            printf("\n");
        }
    }
}
